// http://msdn2.microsoft.com/en-us/library/ms779322(VS.85).aspx
#include <windows.h>
#include <stdio.h>
#include <mmsystem.h>
#include <amstream.h>
#include <includes.h>

/* Notes:
	This class receives audio samples, and plays them out, while keeping track of their PTS (presentation
	time). The class fills up a playback buffer before it actually starts the playback, in order to avoid
	subsequent gaps in the playback. This is needed because audio samples typically arrive in bursts, with
	possibly long periods between bursts.
	It's crucial to keep an accurate count of the current PTS for the audio so the video display can be
	synchronized to it. To do that, we store the AudioPTS of the first sample ever received, and we use
	it, along with the number of played back samples, to compute the current Audio PTS.
	We need to be able to handle situations when the PTS is reset. This occurs in mpeg for example around
	chapters. In such cases, the PTS of the audio we receive might jump back, or forward by an unexpected
	amount (given what we received so far). To deal with that, we detect such events, and reset the 
	reference AudioPTS, when we get to the point, in the playback, where the PTS reset happened.
	The class ONLY WORKS WITH STEREO 48kHZ AUDIO!!!!
*/

class CWaveBuffer {
    public:
        ~CWaveBuffer();
        BOOL Init(HWAVEOUT hWave, int Size);
        BOOL Write(PBYTE pData, int nBytes, int& BytesWritten, bool QueueIt = true);
        void Flush();
        WAVEHDR      m_Hdr;
        HWAVEOUT     m_hWave;
        int          m_nBytes;
		int			 m_ReadyForPlay;	// Indicates buffer is full, ready to play.
};

class CWaveOut {
    public:
        CWaveOut(LPCWAVEFORMATEX Format, int nBuffers, int BufferSize);
        ~CWaveOut();
        void Write(PBYTE Data, int nBytes, int PTS, bool QueueIt = true);
        void Flush();
        void Wait();
        void Reset();
		int  ElapsedBytes();
		void StartPlayback();
        const HANDLE   m_hSem;
        const int      m_nBuffers;
        int            m_CurrentBuffer;
		int			   m_BytesPlayedSoFar;
		int			   m_BufferSize;
        BOOL           m_NoBuffer;
        CWaveBuffer   *m_Hdrs;
        HWAVEOUT       m_hWave;
		bool		   m_OpenedOK;
};

static CWaveOut* MyWaveOut = 0;
static int NumChannels = 0;
static int ReferenceAudioPTS = 0;				// Reference AudioPTS we use to compute the current AudioPTS.
static int NextReferenceAudioPTS = 0;			// Reference AudioPTS that will be used at the next reset.
static int ReferenceNumBytesReceived = 0;		// Number of bytes received corresponding to ReferenceAudioPTS.
static int NextReferenceNumBytesReceived = 0;	// Number of samples received corresponding to ReferenceAudioPTS.
static int NumPreFill = 0;						// How many bytes to fill our queue before starting playback.
static int NumBytesReceived = 0;				// Number of bytes received so far.
static bool DidReset = 0;

extern "C" void closeWaveOut()
{
	delete MyWaveOut; MyWaveOut = NULL;
}	

#define PREFILL_MS 500

#define BYTES_TO_PTS 90/4/48	// Bytes to PTS: /4/48->ms, then *90 -> PTS.
#define BYTES_TO_MS 1/4/48		// Bytes to ms: /4/48->ms.
#define MS_TO_BYTES 48*4		// Milliseconds to bytes.

extern "C" int openWaveOut(float SampleRate, int NumChans)
{
	if(MyWaveOut != 0) delete MyWaveOut; MyWaveOut = NULL;

	WAVEFORMATEX Format;
	ZeroMemory(&Format, sizeof (WAVEFORMATEX));
	Format.wFormatTag = WAVE_FORMAT_PCM;
	Format.wBitsPerSample = 16;
	Format.nChannels = NumChannels = NumChans;
	Format.nSamplesPerSec = (DWORD)SampleRate;
	Format.nBlockAlign = Format.nChannels * Format.wBitsPerSample / 8;
	Format.nAvgBytesPerSec = Format.nSamplesPerSec * Format.nBlockAlign;
	Format.cbSize = 0;
	
	// 6144*8 = 256ms. Multiply by 5e-3 to get ms...
	// I need a large buffer, so I can store a large amount of audio. It's possible that later on,
	// I will need this slack. 600*512 = 1.6s. That's plenty.
	MyWaveOut = new CWaveOut(&Format,600,512);
	ReferenceAudioPTS = 0;
	NextReferenceAudioPTS = 0;
	ReferenceNumBytesReceived = 0;
	NextReferenceNumBytesReceived = 0;
	DidReset = 0;
	NumPreFill = PREFILL_MS * MS_TO_BYTES;

	return MyWaveOut->m_OpenedOK;

}

extern "C" int ElapsedMs()
{
	if(NumChannels && MyWaveOut->m_OpenedOK)
	{
		// BytesPlayedSinceReset is how many bytes have been played since the last reset.
		int BytesPlayedSinceReset = MyWaveOut->ElapsedBytes() - ReferenceNumBytesReceived;
		return (int) BytesPlayedSinceReset * BYTES_TO_MS;
	}
	else
		return 0;

}

extern "C" int GetCurrentAudioPTS(bool* Reset)
{
	*Reset = DidReset;
	DidReset = 0;
	if(!MyWaveOut || !MyWaveOut->m_OpenedOK) return 0;
	return ElapsedMs() * 90  + ReferenceAudioPTS;
}


extern "C" void WriteOut(BYTE* Data, int numBytes, int AudioPTS)
{
	static DWORD thistime = 0;
	static int LastAudioPTS = 0;
	static int NextAudioPTS = 0;

	if(!MyWaveOut->m_OpenedOK) return;

	if(ReferenceAudioPTS == 0) 
	{
		// First time around. Remember the AudioPTS.
		ReferenceAudioPTS = AudioPTS;
		NextReferenceAudioPTS = 0;
		NextAudioPTS = 0;
		LastAudioPTS = 0;
		NumBytesReceived = 0;
		ReferenceNumBytesReceived = 0;
		NextReferenceNumBytesReceived = 0;
	}

	DWORD NewTime = timeGetTime();
//	char Temp[512]; 
	
//	sprintf(Temp,"Time %d ms, Received %d ms\n",NewTime,numBytes/4/48);
//	OutputDebugString(Temp);
	
	// Checking for audio pts discrepancy (current isn't last + number of samples/48*90).
	if(AudioPTS && NextAudioPTS && NextAudioPTS != AudioPTS)
	{
		MYTRACE4("AUDIO PTS DISCREPANCY! Expected %d Current %d Diff %d (%dms)\n",NextAudioPTS, AudioPTS,
			NextAudioPTS-AudioPTS,(NextAudioPTS-AudioPTS)/90);
	}
	// Checking for audio PTS reset (can happen at chapter boundaries). 
	// It's possible for an audio packet not to have a PTS. This could trigger this, of course. We need to
	// avoid that.
	if(AudioPTS)
	{
//		if(LastAudioPTS  && (LastAudioPTS > AudioPTS || NextAudioPTS > AudioPTS + 90000))
		if(LastAudioPTS  && (abs(NextAudioPTS - AudioPTS) > 3000))
		{
			// This is where I should reset.
			MYTRACE4("\nWriteOut AUDIO PTS RESET! Last %d New %d Expected %d BytesPlayedSoFar %d\n",
				LastAudioPTS, AudioPTS, NextAudioPTS,MyWaveOut->m_BytesPlayedSoFar);
			MYTRACE1("NumBytesReceived %d\n\n",NumBytesReceived);
			// When we're done playing NumBytesReceived, we'll reset the PTS to be AudioPTS.
			NextReferenceNumBytesReceived = NumBytesReceived;
			NextReferenceAudioPTS = AudioPTS;
		}
		LastAudioPTS = AudioPTS;
		NextAudioPTS = AudioPTS + numBytes * BYTES_TO_PTS;
	}
	else 
		NextAudioPTS += numBytes * BYTES_TO_PTS;

	// Prefilling code: fill up the audio buffer before starting audio playback.
	if(NumPreFill > 0)
	{
		if(NumBytesReceived == 0)
		{
			// Ramp up first samples to avoid click...
			short* Ptr = (short*) Data;
			float Factor = 1.0f/(min(numBytes/2,2000));
			for(int i=0;i<min(numBytes/2,2000);i++)
				*Ptr++ = (short)(*Ptr * Factor * i);
		}
		// false indicates the data should be written, but not played yet.
		MyWaveOut->Write(Data,numBytes,AudioPTS, false);
		NumPreFill -= numBytes;
		// Buffer has enough samples, start playback.
		if(NumPreFill <= 0) MyWaveOut->StartPlayback();
	}
	else
	{
		// Normal playback. Write and queue samples.
		MyWaveOut->Write(Data,numBytes,AudioPTS, true);
	}
	NumBytesReceived += numBytes;
	thistime = NewTime;
}

extern "C" int GetReadyToPlayMs()
{
#if 1
	int i,j;
	if(MyWaveOut == 0 || !MyWaveOut->m_OpenedOK) return 0;
	// Find out how many of our queued buffers are ready to play. 
	for(i=j=0;i<MyWaveOut->m_nBuffers;i++)
		if(MyWaveOut->m_Hdrs[i].m_Hdr.dwFlags & WHDR_INQUEUE) j++;
	if(j < 10)
	{
//		MYTRACE1("Queue Low: %d buffers \n",j);
	}
	
	return j * MyWaveOut->m_Hdrs[0].m_Hdr.dwBufferLength * BYTES_TO_MS;
#else
	// This is another option, possibly less costly. But it will show a non-zero number of ready
	// samples after the first frame, which we may not want. That's because we need to count the number
	// of samples in the playback queue, not just how many we've received.
	return (NumBytesReceived - MyWaveOut->ElapsedBytes()) * BYTES_TO_MS;
#endif
}

extern "C" int StartPlayback()
{
	if(MyWaveOut && MyWaveOut->m_OpenedOK) MyWaveOut->StartPlayback();
	return 0;
}


BOOL CWaveBuffer::Init(HWAVEOUT hWave, int Size)
{
    m_hWave  = hWave;
    m_nBytes = 0;

    /*  Allocate a buffer and initialize the header. */
    m_Hdr.lpData = (LPSTR)LocalAlloc(LMEM_FIXED, Size);
    if (m_Hdr.lpData == NULL) 
    {
        return FALSE;
    }
    m_Hdr.dwBufferLength  = Size;
    m_Hdr.dwBytesRecorded = 0;
    m_Hdr.dwUser = 0;
    m_Hdr.dwFlags = 0;
    m_Hdr.dwLoops = 0;
    m_Hdr.lpNext = 0;
    m_Hdr.reserved = 0;
	m_ReadyForPlay = 0;

    /*  Prepare it. */
    waveOutPrepareHeader(hWave, &m_Hdr, sizeof(WAVEHDR));
    return TRUE;
}

CWaveBuffer::~CWaveBuffer() 
{
    if (m_Hdr.lpData) 
    {
        waveOutUnprepareHeader(m_hWave, &m_Hdr, sizeof(WAVEHDR));
        LocalFree(m_Hdr.lpData);
    }
}

void CWaveBuffer::Flush()
{
    // ASSERT(m_nBytes != 0);
    m_nBytes = 0;
    waveOutWrite(m_hWave, &m_Hdr, sizeof(WAVEHDR));
}

BOOL CWaveBuffer::Write(PBYTE pData, int nBytes, int& BytesWritten, bool QueueIt)
{
    // ASSERT((DWORD)m_nBytes != m_Hdr.dwBufferLength);
    BytesWritten = min((int)m_Hdr.dwBufferLength - m_nBytes, nBytes);
    CopyMemory((PVOID)(m_Hdr.lpData + m_nBytes), (PVOID)pData, BytesWritten);
    m_nBytes += BytesWritten;
    if (m_nBytes == (int)m_Hdr.dwBufferLength) 
    {
        /*  Write it! */
        m_nBytes = 0;
		m_ReadyForPlay = true;
		// If queue it is true, play it. Otherwise, simply return.
		if(QueueIt) waveOutWrite(m_hWave, &m_Hdr, sizeof(WAVEHDR));
        return TRUE;
    }
    return FALSE;
}

void CALLBACK WaveCallback(HWAVEOUT hWave, UINT uMsg, DWORD dwUser, 
                           DWORD dw1, DWORD dw2)
{
    if (uMsg == WOM_DONE) 
    {
		// Release semaphore in case the write thread is waiting for a buffer.
		CWaveOut* TheWaveOut = (CWaveOut*)dwUser;
        ReleaseSemaphore(TheWaveOut->m_hSem, 1, NULL);
		// Now's the time to check whether the PTS needs to be reset. To do that we compare
		// m_BytesPlayedSoFar (how many bytes have been played back) to NextReferenceNumBytesReceived
		// (number of played-back bytes where PTS reset should happen).
		if(NextReferenceNumBytesReceived && 
			TheWaveOut->m_BytesPlayedSoFar >= NextReferenceNumBytesReceived)
		{
			// PTS RESET. Simply, set 
			MYTRACE1("\nCallback pts RESET! BytesPlayedSoFar %d\n\n",TheWaveOut->m_BytesPlayedSoFar);
			ReferenceNumBytesReceived = NextReferenceNumBytesReceived;
			ReferenceAudioPTS = NextReferenceAudioPTS;
			NextReferenceNumBytesReceived = 0;
			DidReset = true;
		}
		TheWaveOut->m_BytesPlayedSoFar += TheWaveOut->m_BufferSize;
    }
}

CWaveOut::CWaveOut(LPCWAVEFORMATEX Format, int nBuffers, int BufferSize) :
    m_nBuffers(nBuffers),
    m_CurrentBuffer(0),
    m_NoBuffer(TRUE),
    m_hSem(CreateSemaphore(NULL, nBuffers, nBuffers, NULL)),
    m_Hdrs(new CWaveBuffer[nBuffers]),
    m_hWave(NULL)
{
    /*  Create wave device. */
    MMRESULT Ret = waveOutOpen(&m_hWave,
                WAVE_MAPPER,
                Format,
                (DWORD)WaveCallback,
 //               (DWORD)m_hSem,
                (DWORD)this,
                CALLBACK_FUNCTION);

    /*  Initialize the wave buffers. */
	m_BytesPlayedSoFar = 0;
	m_BufferSize = BufferSize;
	m_OpenedOK = (Ret == MMSYSERR_NOERROR);
	if(!m_OpenedOK) return;
   for (int i = 0; i < nBuffers; i++) 
        m_Hdrs[i].Init(m_hWave, BufferSize);
}

CWaveOut::~CWaveOut()
{
    /*  First, get the buffers back. */
    waveOutReset(m_hWave);
    /*  Free the buffers. */
    delete [] m_Hdrs;
    /*  Close the wave device. */
    waveOutClose(m_hWave);
    /*  Free the semaphore. */
    CloseHandle(m_hSem);
}

void CWaveOut::Flush()
{
    if (!m_NoBuffer) 
    {
        m_Hdrs[m_CurrentBuffer].Flush();
        m_NoBuffer = TRUE;
        m_CurrentBuffer = (m_CurrentBuffer + 1) % m_nBuffers;
    }
}

int CWaveOut::ElapsedBytes()
{
	MMTIME Time;
	Time.wType = TIME_BYTES;
	// This will return how many bytes we've played so far.
	waveOutGetPosition(m_hWave,&Time,sizeof(Time));
	return Time.u.sample;
}


void CWaveOut::Reset()
{
    waveOutReset(m_hWave);
}

void CWaveOut::Write(PBYTE pData, int nBytes, int PTS, bool QueueIt)
{
	int StartnBytes = nBytes;
    while (nBytes != 0) 
    {
        /*  Get a buffer if necessary. */
        if (m_NoBuffer) 
        {
			// Test whether the next buffer is available. 
//			if((m_Hdrs[(m_CurrentBuffer + 1) % m_nBuffers].m_Hdr.dwFlags & WHDR_INQUEUE))
//				OutputDebugString("BUFFER WAIT\n");
            WaitForSingleObject(m_hSem, INFINITE);
            m_NoBuffer = FALSE;
        }

        /*  Write into a buffer. */
        int nWritten;
        if (m_Hdrs[m_CurrentBuffer].Write(pData, nBytes, nWritten, QueueIt)) 
        {
            m_NoBuffer = TRUE;
            m_CurrentBuffer = (m_CurrentBuffer + 1) % m_nBuffers;
            nBytes -= nWritten;
            pData += nWritten;
        } 
        else 
        {
            // ASSERT(nWritten == nBytes);
            break;
        }
    }
}

void CWaveOut::Wait()
{
    /*  Send any remaining buffers. */
    Flush();
    /*  Wait for the buffers back. */
    for (int i = 0; i < m_nBuffers; i++) 
    {
        WaitForSingleObject(m_hSem, INFINITE);
    }
    LONG lPrevCount;
    ReleaseSemaphore(m_hSem, m_nBuffers, &lPrevCount);
}

void CWaveOut::StartPlayback()
{
//	char Temp[512];
	MYTRACE1("Starting Playback Prefill %d\n",NumPreFill);
	// Go through all buffers and send all the ones that are ready to the wave out device.
    for (int i = 0; i < m_nBuffers; i++) 
    {
 		if(!MyWaveOut->m_Hdrs[i].m_ReadyForPlay) break;
		waveOutWrite(MyWaveOut->m_Hdrs[i].m_hWave, &MyWaveOut->m_Hdrs[i].m_Hdr, sizeof(WAVEHDR));
    }
	NumPreFill = 0;
}
